/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "opengl.pmh"

namespace fe
{

#define DGL_VERTEXARRAY		TRUE	// allow render calls in group form
#define DGL_LINEWIDTH		TRUE	// allow custom linewidth
#define DGL_ANTIALIAS_LINE	TRUE	// allow antialiased lines
#define DGL_POINT_ON_LINE	FALSE	// make sure short lines show up
#define DGL_THREAD_LOCKING	TRUE	// lock the event context

#define DGL_CHECK_ERROR		(FE_CODEGEN<FE_DEBUG)	// check glGetError
#define	DGL_CHECK_ALPHA		(FE_CODEGEN<=FE_DEBUG)	// assert if alpha<=0.0

#if FE_DOUBLE_REAL
#define	DGL_REAL			GL_DOUBLE
#else
#define	DGL_REAL			GL_FLOAT
#endif

DrawOpenGL::DrawOpenGL(void):
	m_fontbase(0),
	m_isDirect(TRUE),
	m_lockThreads(TRUE)
{
#if FE_OS==FE_LINUX
	Display *pDisplay=glXGetCurrentDisplay();
	GLXContext glxContext=glXGetCurrentContext();
	m_isDirect=glXIsDirect(pDisplay,glxContext);
//	feLog("pDisplay=%p glxContext=%d direct=%d\n",
//			pDisplay,glxContext,m_isDirect);
#endif

	if(!m_isDirect)
	{
		feLog("DrawOpenGL: indirect context detected ->"
				" simplifying default style\n");
	}

	fe::String value;
	if(!fe::System::getEnvironmentVariable("FE_GL_LOCKING", value) ||
			!atoi(value.c_str()))
	{
		m_lockThreads=FALSE;
	}

	m_pQuadric=gluNewQuadric();
}

DrawOpenGL::~DrawOpenGL()
{
	gluDeleteQuadric(m_pQuadric);
}

void DrawOpenGL::initialize(void)
{
	BWORD found_context=FALSE;
#if FE_2DGL==FE_2D_X_GFX	// Linux or X11-mode OSX
	found_context=(glXGetCurrentContext()!=NULL);
#elif FE_OS==FE_WIN32 || FE_OS==FE_WIN64
	found_context=(wglGetCurrentContext()!=NULL);
#endif

	if(!found_context)
	{
		feX(e_cannotCreate,"DrawOpenGL::DrawOpenGL()","no GL context");
	}
	else
	{
		glEnable(GL_DEPTH_TEST);

		createDefaultTexture();
		useDefaultTexture();
		setupLights();
	}
}

void DrawOpenGL::createDefaultView(void)
{
	m_spDefaultView=registry()->create("ViewI.ViewOpenGL");
#if FALSE
	if(!m_spDefaultView.isValid())
	{
		feX(e_cannotCreate,"DrawOpenGL::initialize",
				"failed to create a ViewI.ViewOpenGL");
	}
#endif
}

void DrawOpenGL::threadLock(void)
{
	sp<WindowI> spWindowI=view()->window();
	if(spWindowI.isValid())
	{
		sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
		if(spEventContextI.isValid())
		{
			spEventContextI->threadLock();
		}
	}
}

void DrawOpenGL::threadUnlock(void)
{
	sp<WindowI> spWindowI=view()->window();
	if(spWindowI.isValid())
	{
		sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
		if(spEventContextI.isValid())
		{
			spEventContextI->threadUnlock();
		}
	}
}

void DrawOpenGL::flush(void)
{
	sp<WindowI> spWindowI=view()->window();
	spWindowI->makeCurrent();
	spWindowI->swapBuffers();
	spWindowI->clear();
}

void DrawOpenGL::checkForError(void)
{
	GLenum error=glGetError();
	if(error != GL_NO_ERROR)
	{
		feLog("DrawOpenGL::checkForError glGetError 0x%x\n",error);
	}
	FEASSERT(error == GL_NO_ERROR);
}

void DrawOpenGL::setDrawMode(sp<DrawMode> spMode)
{
	SAFEGUARD;
	DrawCommon::setDrawMode(spMode);

	DrawMode::DrawStyle style=drawStyle();

	const BWORD lit=isLit();
	const BWORD texture=(drawStyle()==DrawMode::e_solid);

	setup(style,lit,twoSidedLighting(),texture,backfaceCulling());
	glPointSize(pointSize());

	bindTexture(textureImage());
}

void DrawOpenGL::pushAntialiasPoints(void)
{
	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT |
			GL_ENABLE_BIT | GL_HINT_BIT | GL_LIGHTING_BIT);

//	glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_BLEND);
	glShadeModel(GL_FLAT);
	glDepthMask(GL_FALSE);
}

void DrawOpenGL::pushAntialiasLines(void)
{
	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT |
			GL_ENABLE_BIT | GL_HINT_BIT | GL_LIGHTING_BIT);

//	glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glShadeModel(GL_FLAT);
	glDepthMask(GL_FALSE);
}

void DrawOpenGL::pushBlending(void)
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_BLEND);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void DrawOpenGL::drawPoints(const SpatialVector *vertex,
		const SpatialVector *normal,U32 vertices,
		BWORD multicolor,const Color *color)
{
	FEASSERT(vertex);
	FEASSERT(vertices > 0);

	const BWORD lit=isLit();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	const BWORD normals=normal && lit;
	const BWORD toggle_lighting=!normal && lit;
	if(toggle_lighting)
	{
		glDisable(GL_LIGHTING);
	}

	pushAntialiasPoints();

#if DGL_VERTEXARRAY

	if (multicolor && color)
	{
		glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
		glEnableClientState(GL_VERTEX_ARRAY);

		glColorPointer(4,DGL_REAL,sizeof(Color),color);
		glEnableClientState(GL_COLOR_ARRAY);

		if(normals)
		{
			glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
			glEnableClientState(GL_NORMAL_ARRAY);
		}

		glDrawArrays(GL_POINTS,0,vertices);

		if(normals)
		{
			glDisableClientState(GL_NORMAL_ARRAY);
		}

		glDisableClientState(GL_COLOR_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
	}
	else
	{
		if (color)
		{
#if FE_DOUBLE_REAL
			glColor4dv(color[0].temp());
#else
			glColor4fv(color[0].temp());
#endif
		}

		glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
		glEnableClientState(GL_VERTEX_ARRAY);

		if(normals)
		{
			glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
			glEnableClientState(GL_NORMAL_ARRAY);
		}

		glDrawArrays(GL_POINTS,0,vertices);

		if(normals)
		{
			glDisableClientState(GL_NORMAL_ARRAY);
		}

		glDisableClientState(GL_VERTEX_ARRAY);
	}
#else
	glBegin(GL_POINTS);

	if(normals)
	{
		U32 m;
		if(multicolor && color)
		{
			for(m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
				glColor4dv(color[m].temp());
				glNormal3dv(normal[m].temp());
				glVertex3dv(vertex[m].temp());
#else
				glColor4fv(color[m].temp());
				glNormal3fv(normal[m].temp());
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
		else
		{
			if (color)
			{
#if FE_DOUBLE_REAL
				glColor4dv(color[0].temp());
#else
				glColor4fv(color[0].temp());
#endif
			}
			for (m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
					glNormal3dv(normal[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glNormal3fv(normal[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
			}
		}
	}
	else
	{
		U32 m;
		if(multicolor && color)
		{
			for(m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
				glColor4dv(color[m].temp());
				glVertex3dv(vertex[m].temp());
#else
				glColor4fv(color[m].temp());
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
		else
		{
			if (color)
			{
#if FE_DOUBLE_REAL
				glColor4dv(color[0].temp());
#else
				glColor4fv(color[0].temp());
#endif
			}
			for (m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
				glVertex3dv(vertex[m].temp());
#else
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
	}

	glEnd();
#endif

	//* restore point smooth
	glPopAttrib();

	if(toggle_lighting)
	{
		glEnable(GL_LIGHTING);
	}

#if DGL_CHECK_ERROR
	checkForError();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawLines(const SpatialVector* vertex,
		const SpatialVector *normal,U32 vertices,
		StripMode strip,BWORD multicolor,const Color* color)
{
	if(drawStyle()==DrawMode::e_pointCloud)
	{
		drawPoints(vertex,NULL,vertices,multicolor,color);
		return;
	}

	FEASSERT(vertex);
	FEASSERT(vertices > 1);
	FEASSERT(strip != e_fan);

#if DGL_CHECK_ALPHA
	if(!multicolor && color && (*color)[3]<=0.0f)
	{
		feX("DrawOpenGL::drawLines",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(*color));
	}
#endif

	const BWORD lit=isLit();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

/*
	feLog("drawLines normal %p lit %d aa %d width %.6G"
			" multicolor %d color %p strip %d lock %d"
			" projection %d window %d viewmode %d\n",
			normal,lit,antialias(),lineWidth(),
			multicolor,color,strip,m_lockThreads,
			view()->projection(),view()->window().isValid(),
			view()->viewMode().isValid());
*/

	const BWORD normals=normal && lit;
	const BWORD toggle_lighting=!normal && lit;
	if(toggle_lighting)
	{
		glDisable(GL_LIGHTING);
	}

#if DGL_ANTIALIAS_LINE
	if(antialias())
		pushAntialiasLines();
#endif

#if DGL_LINEWIDTH
	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	if(pushLineWidth)
	{
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(lineWidth());
	}
#endif

#if DGL_VERTEXARRAY
	if (multicolor && color)
	{
		glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
		glEnableClientState(GL_VERTEX_ARRAY);

		glColorPointer(4,DGL_REAL,sizeof(Color),color);
		glEnableClientState(GL_COLOR_ARRAY);

		if(normals)
		{
			glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
			glEnableClientState(GL_NORMAL_ARRAY);
		}

		glDrawArrays(strip==e_strip? GL_LINE_STRIP: GL_LINES,0,vertices);

		if(normals)
		{
			glDisableClientState(GL_NORMAL_ARRAY);
		}
		glDisableClientState(GL_COLOR_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
	}
	else
	{
		if(color)
		{
#if FE_DOUBLE_REAL
			glColor4dv(color[0].temp());
#else
			glColor4fv(color[0].temp());
#endif
		}

		glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
		glEnableClientState(GL_VERTEX_ARRAY);

		if(normals)
		{
			glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
			glEnableClientState(GL_NORMAL_ARRAY);
		}

		glDrawArrays(strip==e_strip? GL_LINE_STRIP: GL_LINES,0,vertices);

		if(normals)
		{
			glDisableClientState(GL_NORMAL_ARRAY);
		}
		glDisableClientState(GL_VERTEX_ARRAY);
	}
#else
	glBegin(strip==e_strip? GL_LINE_STRIP: GL_LINES);

	U32 m;

	if(normals)
	{
		if (multicolor && color)
		{
			for (m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
				glColor4dv(color[m].temp());
				glNormal3dv(normal[m].temp());
				glVertex3dv(vertex[m].temp());
#else
				glColor4fv(color[m].temp());
				glNormal3fv(normal[m].temp());
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
		else
		{
			if (color)
#if FE_DOUBLE_REAL
				glColor4dv(color[0].temp());
#else
				glColor4fv(color[0].temp());
#endif
			for (m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
				glNormal3dv(normal[m].temp());
				glVertex3dv(vertex[m].temp());
#else
				glNormal3fv(normal[m].temp());
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
	}
	else
	{
		if (multicolor && color)
		{
			for (m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
				glColor4dv(color[m].temp());
				glVertex3dv(vertex[m].temp());
#else
				glColor4fv(color[m].temp());
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
		else
		{
			if (color)
#if FE_DOUBLE_REAL
				glColor4dv(color[0].temp());
#else
				glColor4fv(color[0].temp());
#endif
			for (m=0; m < vertices; m++)
#if FE_DOUBLE_REAL
				glVertex3dv(vertex[m].temp());
#else
				glVertex3fv(vertex[m].temp());
#endif
		}
	}

	glEnd();
#endif

#if DGL_POINT_ON_LINE
	// make sure it shows up when only a point
#if FE_DOUBLE_REAL
	glColor4dv(color[0].temp());
#else
	glColor4fv(color[0].temp());
#endif
	glBegin(GL_POINTS);
#if FE_DOUBLE_REAL
	glVertex3dv(vertex[0].temp());
#else
	glVertex3fv(vertex[0].temp());
#endif
	glEnd();
#endif

#if DGL_LINEWIDTH
	if(pushLineWidth)
	{
		//* restore line width
		glPopAttrib();
	}
#endif

#if DGL_ANTIALIAS_LINE
	if(antialias())
		glPopAttrib();
#endif

	if(toggle_lighting)
	{
		glEnable(GL_LIGHTING);
	}

#if DGL_CHECK_ERROR
	checkForError();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawTriangles(const SpatialVector* vertex,
		const SpatialVector* normal,const Vector2f* texture,U32 vertices,
		StripMode strip,BWORD multicolor,const Color* color)
{
	if(drawStyle()==DrawMode::e_wireframe ||
			drawStyle()==DrawMode::e_pointCloud)
	{
		if(strip==e_fan)
		{
			//* not exactly...
			strip=e_strip;
		}
		drawLines(vertex,normal,vertices,strip,multicolor,color);
	}
	else
	{
		drawPolys(3,vertex,normal,texture,vertices,strip,multicolor,color);
	}
}

void DrawOpenGL::drawRectangles(const SpatialVector *vertex,
		U32 vertices,BWORD multicolor,const Color *color)
{
	drawPolys(2,vertex,NULL,NULL,vertices,e_discrete,multicolor,color);
}

//* private: not generalized; only well defined in how it is called internally
void DrawOpenGL::drawPolys(U32 grain,const SpatialVector* vertex,
		const SpatialVector* normal,const Vector2f* texture,U32 vertices,
		StripMode strip,BWORD multicolor,const Color* color)
{
#if FALSE
	feLog(" DrawOpenGL::drawPolys %d <%s> <%s> %d %d %d <%s>\n",
			grain,
			vertex? print(vertex[0]).c_str(): "NULL",
			normal? print(normal[0]).c_str(): "NULL",
			vertices,(U32)strip,multicolor,
			color? print(*color).c_str(): "NULL");
#endif
	FEASSERT(grain>1);
	FEASSERT(vertices>=grain);
	FEASSERT(strip!=e_discrete || !(vertices%grain));
	FEASSERT(vertex);
	FEASSERT(drawStyle()==DrawMode::e_outline || normal || grain<3);

#if DGL_CHECK_ALPHA
	if(!multicolor && color && (*color)[3]<=0.0f)
	{
		feX("DrawOpenGL::drawPolys",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(*color));
	}
#endif

	GLenum primitive=GL_QUADS;
	if(grain==3)
	{
		primitive=(strip==e_discrete)? GL_TRIANGLES:
				((strip==e_strip)? GL_TRIANGLE_STRIP:
				GL_TRIANGLE_FAN);
	}

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	const Real alpha=color? color[0][3]: 1.0f;
	const BWORD lit=isLit();
	const BWORD blending=((multicolor || alpha!=1.0f) && lit);
	if(blending)
	{
		pushBlending();
	}

	U32 passes=1+(drawStyle()==DrawMode::e_edgedSolid);
	for(U32 pass=0;pass<passes;pass++)
	{
		if(pass)
		{
#if DGL_LINEWIDTH
			if(pushLineWidth)
			{
				glPushAttrib(GL_LINE_BIT);
				glLineWidth(lineWidth());
			}
#endif
			setup(DrawMode::e_outline,FALSE,FALSE,FALSE,FALSE);
			glEnable(GL_POLYGON_OFFSET_LINE);
			glPolygonOffset(-1.0f, 0.0f);
		}

#if DGL_VERTEXARRAY
		const BWORD normals=!pass && lit;

		if (multicolor && color)
		{
			glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
			glEnableClientState(GL_VERTEX_ARRAY);

			glColorPointer(4,DGL_REAL,sizeof(Color),color);
			glEnableClientState(GL_COLOR_ARRAY);

			if(normals)
			{
				glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
				glEnableClientState(GL_NORMAL_ARRAY);
			}
			if(texture)
			{
				glTexCoordPointer(2,GL_FLOAT,sizeof(Vector2f),texture);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			}

			glDrawArrays(primitive,0,vertices);

			if(texture)
			{
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			}
			if(normals)
			{
				glDisableClientState(GL_NORMAL_ARRAY);
			}

			glDisableClientState(GL_COLOR_ARRAY);
			glDisableClientState(GL_VERTEX_ARRAY);
		}
		else
		{
			if(color)
			{
				if(blending)
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				else
#if FE_DOUBLE_REAL
					glColor3dv(color[0].temp());
#else
					glColor3fv(color[0].temp());
#endif
			}

			if(grain>2)
			{
				glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
				glEnableClientState(GL_VERTEX_ARRAY);

				if(normals)
				{
					glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
					glEnableClientState(GL_NORMAL_ARRAY);
				}
				if(texture)
				{
					glTexCoordPointer(2,GL_FLOAT,sizeof(Vector2f),texture);
					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				}

				glDrawArrays(primitive,0,vertices);

				if(texture)
				{
					glDisableClientState(GL_TEXTURE_COORD_ARRAY);
				}
				if(normals)
				{
					glDisableClientState(GL_NORMAL_ARRAY);
				}
				glDisableClientState(GL_VERTEX_ARRAY);
			}
			else
			{
				glBegin(primitive);
				for(U32 m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glVertex3d(vertex[m].temp()[0],vertex[m+1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#else
					glVertex3f(vertex[m].temp()[0],vertex[m+1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#endif

					m++;
#if FE_DOUBLE_REAL
					glVertex3d(vertex[m].temp()[0],vertex[m-1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#else
					glVertex3f(vertex[m].temp()[0],vertex[m-1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#endif
				}
				glEnd();
			}
		}
#else
		glBegin(primitive);

		if(multicolor && color)
		{
			for(U32 m=0; m < vertices; m++)
			{
				if (color[m][3] != 1.0f)
#if FE_DOUBLE_REAL
					glColor4dv(color[m].temp());
#else
					glColor4fv(color[m].temp());
#endif
				else
#if FE_DOUBLE_REAL
					glColor3dv(color[m].temp());
#else
					glColor3fv(color[m].temp());
#endif

#if FE_DOUBLE_REAL
				glNormal3dv(normal[m].temp());
				glVertex3dv(vertex[m].temp());
#else
				glNormal3fv(normal[m].temp());
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
		else
		{
			if(color)
			{
				if(blending)
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				else
#if FE_DOUBLE_REAL
					glColor3dv(color[0].temp());
#else
					glColor3fv(color[0].temp());
#endif
			}

			if(grain>2)
			{
				for(U32 m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glNormal3dv(normal[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glNormal3fv(normal[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
			else
			{
				for(U32 m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glVertex3d(vertex[m].temp()[0],vertex[m+1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#else
					glVertex3f(vertex[m].temp()[0],vertex[m+1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#endif

					m++;
#if FE_DOUBLE_REAL
					glVertex3d(vertex[m].temp()[0],vertex[m-1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#else
					glVertex3f(vertex[m].temp()[0],vertex[m-1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#endif
				}
			}
		}

		glEnd();
#endif

		if(pass)
		{
			setup(drawStyle(),lit,twoSidedLighting(),FALSE,backfaceCulling());
			glDisable(GL_POLYGON_OFFSET_LINE);

#if DGL_LINEWIDTH
			if(pushLineWidth)
			{
				//* restore line width
				glPopAttrib();
			}
#endif
		}
	}

	if(blending)
	{
		glPopAttrib();
	}

#if DGL_CHECK_ERROR
	checkForError();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawAlignedText(const SpatialVector &location,
		const String text,const Color &color)
{
	const BWORD lit=isLit();

	if(!m_fontbase)
	{
		feLog("DrawOpenGL::drawAlignedText fontbase==0\n");
		return;
	}

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	if(lit)
	{
		glDisable(GL_LIGHTING);
	}

#if FE_DOUBLE_REAL
	glColor3dv(color.temp());
	glRasterPos3dv(location.temp());
#else
	glColor3fv(color.temp());
	glRasterPos3fv(location.temp());
#endif

	glPushAttrib(GL_LIST_BIT);

	glListBase(m_fontbase);
	glCallLists(text.length(),GL_UNSIGNED_BYTE,(unsigned char *)text.c_str());

	glPopAttrib();

	if(lit)
	{
		glEnable(GL_LIGHTING);
	}

//	glFinish();

#if DGL_CHECK_ERROR
	checkForError();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawSphereGLU(const SpatialTransform &transform,
		const SpatialVector *scale,const Color &color)
{
	const GLint slices=8;	// 6
	const GLint stacks=6;	// 4

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

#if DGL_ANTIALIAS_LINE
	BWORD antialiasAll=antialias() && drawStyle()<DrawMode::e_solid;
	if(antialiasAll)
		pushAntialiasLines();
#endif

#if DGL_LINEWIDTH
	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	if(pushLineWidth)
	{
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(lineWidth());
	}
#endif

#if DGL_CHECK_ALPHA
	if(color[3]<=0.0f)
	{
		feX("DrawOpenGL::drawSphereGLU",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(color));
	}
#endif

	const Real alpha=color[3];
	if (alpha != 1.0f)
	{
		pushBlending();
	}
	glPushMatrix();
	Real raw[16];
	transform.copy16(raw);
#if FE_DOUBLE_REAL
	glMultMatrixd(raw);
#else
	glMultMatrixf(raw);
#endif

#if FE_DOUBLE_REAL
	glColor4dv(color.temp());
#else
	glColor4fv(color.temp());
#endif

	//* TODO scale appropriately

	gluSphere(m_pQuadric,(*scale)[0],slices,stacks);
	if(drawStyle()==DrawMode::e_edgedSolid)
	{
#if DGL_ANTIALIAS_LINE
		if(antialias())
			pushAntialiasLines();
#endif

		setup(DrawMode::e_outline,FALSE,twoSidedLighting(),FALSE,
				backfaceCulling());
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(-1.0f, 0.0f);

		gluSphere(m_pQuadric,(*scale)[0],slices,stacks);

		setup(drawStyle(),TRUE,twoSidedLighting(),FALSE,backfaceCulling());
		glDisable(GL_POLYGON_OFFSET_LINE);

#if DGL_ANTIALIAS_LINE
		if(antialias())
			glPopAttrib();
#endif
	}

	glPopMatrix();

	if (alpha != 1.0f)
	{
		glPopAttrib();
	}

#if DGL_LINEWIDTH
	if(pushLineWidth)
	{
		//* restore line width
		glPopAttrib();
	}
#endif

#if DGL_ANTIALIAS_LINE
	if(antialiasAll)
		glPopAttrib();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawCylinderGLU(const SpatialTransform &transform,
		const SpatialVector *scale,Real baseScale,const Color &color,
		U32 slices)
{
	const U32 stacks=1;

	if(slices<1)
	{
		slices=6;	// tweak
	}
	else if(slices<3)
	{
		slices=3;
	}

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

#if DGL_ANTIALIAS_LINE
	BWORD antialiasAll=antialias() && drawStyle()<DrawMode::e_solid;
	if(antialiasAll)
		pushAntialiasLines();
#endif
#if DGL_LINEWIDTH
	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	if(pushLineWidth)
	{
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(lineWidth());
	}
#endif

#if DGL_CHECK_ALPHA
	if(color[3]<=0.0f)
	{
		feX("DrawOpenGL::drawCylinderGLU",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(color));
	}
#endif

	const Real alpha=color[3];
	const BWORD lit=isLit();
	const BWORD blending=(alpha!=1.0f && lit);
	if(blending)
	{
		pushBlending();
	}

	glPushMatrix();
	Real raw[16];
	transform.copy16(raw);
#if FE_DOUBLE_REAL
	glMultMatrixd(raw);
#else
	glMultMatrixf(raw);
#endif

	if(blending)
	{
#if FE_DOUBLE_REAL
		glColor4dv(color.temp());
#else
		glColor4fv(color.temp());
#endif
	}
	else
	{
#if FE_DOUBLE_REAL
		glColor3dv(color.temp());
#else
		glColor3fv(color.temp());
#endif
	}

	//* TODO scale appropriately

	gluCylinder(m_pQuadric,
			(*scale)[0]*baseScale,	//* base radius
			(*scale)[0],			//* top radius
			(*scale)[2],			//* height
			slices,					//* around
			stacks);				//* along

	if(drawStyle()==DrawMode::e_edgedSolid)
	{
#if DGL_ANTIALIAS_LINE
		if(antialias())
			pushAntialiasLines();
#endif

		setup(DrawMode::e_outline,FALSE,twoSidedLighting(),FALSE,
				backfaceCulling());
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(-1.0f, 0.0f);

		gluCylinder(m_pQuadric,
				(*scale)[0]*baseScale,	//* base radius
				(*scale)[0],			//* top radius
				(*scale)[2],			//* height
				slices,					//* around
				stacks);				//* along

		setup(drawStyle(),TRUE,twoSidedLighting(),FALSE,backfaceCulling());
		glDisable(GL_POLYGON_OFFSET_LINE);

#if DGL_ANTIALIAS_LINE
		if(antialias())
			glPopAttrib();
#endif
	}

	glPopMatrix();

	if(blending)
	{
		glPopAttrib();
	}

#if DGL_LINEWIDTH
	if(pushLineWidth)
	{
		//* restore line width
		glPopAttrib();
	}
#endif

#if DGL_ANTIALIAS_LINE
	if(antialiasAll)
		glPopAttrib();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawTransformedTriangles(const SpatialTransform &transform,
		const SpatialVector *scale,const SpatialVector *vertex,
		const SpatialVector *normal,const Vector2f *texture,U32 vertices,
		StripMode strip,BWORD multicolor,const Color *color)
{
	Real matrix[16];
	transform.copy16(matrix);

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	glPushMatrix();
#if FE_DOUBLE_REAL
	glMultMatrixd(matrix);
#else
	glMultMatrixf(matrix);
#endif

	if(scale)
	{
		glScalef((*scale)[0],(*scale)[1],(*scale)[2]);
	}

	drawTriangles(vertex,normal,texture,vertices,strip,multicolor,color);
	glPopMatrix();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::setupLights(void)
{
	const GLfloat position[]={ 1.0f, -1.0f, 3.0f, 0.0f };
	const GLfloat ambient[]={ 0.1f, 0.1f, 0.1f, 0.1f };
	const GLfloat diffuse[]={ 0.5f, 0.5f, 0.5f, 0.5f };
	const GLfloat specular[]={ 1.0f, 1.0f, 1.0f, 1.0f };

	glLightfv(GL_LIGHT0,GL_POSITION,position);
	glLightfv(GL_LIGHT0,GL_AMBIENT,ambient);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);
	glLightfv(GL_LIGHT0,GL_SPECULAR,specular);

	// glScalef will affect lighting unless auto-normalized
	glEnable(GL_NORMALIZE);
}

void DrawOpenGL::setup(DrawMode::DrawStyle style,BWORD lit,BWORD dual,
		BWORD texture,BWORD cull)
{
	gluQuadricDrawStyle(m_pQuadric,
			drawStyle()==DrawMode::e_wireframe? GLU_SILHOUETTE:
			(drawStyle()==DrawMode::e_pointCloud? GLU_POINT: GLU_FILL));
	gluQuadricTexture(m_pQuadric,texture? GL_TRUE: GL_FALSE);

	if(style==DrawMode::e_outline)
	{
		glPolygonMode(GL_FRONT,GL_LINE);
	}

	if(lit)
	{
		if(texture)
		{
			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
//			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
//			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);

//			glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
			glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

			glEnable(GL_TEXTURE_2D);
//			glShadeModel(GL_FLAT);
		}
		else
		{
			glDisable(GL_TEXTURE_2D);
		}

#if FALSE
		GLfloat colormap[]={ 11.0f, 159.0f, 159.0f };
		glMaterialfv(GL_FRONT,GL_COLOR_INDEXES,colormap);
#endif

//		const GLfloat alpha=0.7f;

#if FALSE
		const GLfloat specular[]={ 0.0f, 0.0f, 0.0f, alpha };
		const GLfloat shininess[]={ 10.0f };

		glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,specular);
		glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shininess);
#endif

#if FALSE
		const GLfloat black[]={ 0.0f, 0.0f, 0.0f, 1.0f };
		const GLfloat diffuse0[]={ 0.4f, 0.4f, 0.4f, alpha };

		const GLfloat ambient1[]={ 0.0f, 0.4f, 0.0f, alpha };
		const GLfloat diffuse1[]={ 0.0f, 0.6f, 0.0f, alpha };

		const GLfloat ambient2[]={ 0.0f, 0.0f, 0.4f, alpha };
		const GLfloat diffuse2[]={ 0.0f, 0.0f, 0.6f, alpha };

		if(dual)
		{
			glMaterialfv(GL_BACK,GL_AMBIENT,ambient2);
			glMaterialfv(GL_BACK,GL_DIFFUSE,diffuse2);
		}
		else
		{
			glMaterialfv(GL_BACK,GL_AMBIENT,ambient1);
			glMaterialfv(GL_BACK,GL_DIFFUSE,diffuse1);
		}

		if(style==DrawMode::e_edgedSolid)
		{
			glMaterialfv(GL_FRONT,GL_AMBIENT,black);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse0);
		}
		else
		{
			glMaterialfv(GL_FRONT,GL_AMBIENT,ambient1);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse1);
		}
		glDisable(GL_COLOR_MATERIAL);
#else
		//* follow glColor calls
		glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
#endif

		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glPolygonMode(GL_FRONT,GL_FILL);
		glShadeModel(GL_SMOOTH);
	}
	else
	{
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
	}

	//* Backface Culling
	if(cull || style==DrawMode::e_outline)
	{
		glEnable(GL_CULL_FACE);
	}
	else
	{
		glDisable(GL_CULL_FACE);
	}

	//* Two Sided Lighting
	if(cull || style==DrawMode::e_outline || !dual)
	{
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
	}
	else
	{
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
	}
}

void DrawOpenGL::createDefaultTexture(void)
{
	const long bit=0x8;
	unsigned char c;
	long m,n;
	for(m=0;m<64;m++)
	{
		for(n=0;n<64;n++)
		{
			c=128+127*(((m&bit)==0)^(((n&bit))==0));
			m_texture[m][n][0]=383-c;
			m_texture[m][n][1]=c;
			m_texture[m][n][2]=c;
		}
	}

	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
}

void DrawOpenGL::useDefaultTexture(void)
{
	glTexImage2D(GL_TEXTURE_2D,0,3,
			64,64,0,GL_RGB,GL_UNSIGNED_BYTE,
			m_texture);
}

void DrawOpenGL::queryFormat(sp<ImageI> spImageI,GLenum& format,
		GLenum& type,GLint& components)
{
	switch(spImageI->format())
	{
		case ImageI::e_colorindex:
			format=GL_COLOR_INDEX;
			type=GL_UNSIGNED_SHORT;
			components=1;
			break;
		case ImageI::e_rgba:
			format=GL_RGBA;
			type=GL_UNSIGNED_BYTE;
			components=4;
			break;
		default:
			format=GL_RGB;
			type=GL_UNSIGNED_BYTE;
			components=3;
			break;
	}
}

void DrawOpenGL::drawRaster(sp<ImageI> spImageI,const SpatialVector& location)
{
#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

//	glPushAttrib(GL_DEPTH_BUFFER_BIT);
//	glDisable(GL_DEPTH_TEST);

	glRasterPos3f((GLfloat)location[0],(GLfloat)location[1],
			(GLfloat)location[2]);
	glPixelZoom(1.0f,-1.0f);

	GLenum format;
	GLenum type;
	GLint components;
	queryFormat(spImageI,format,type,components);

	glDrawPixels((GLsizei)spImageI->width(),(GLsizei)spImageI->height(),
			format,type,(const GLvoid*)spImageI->raw());

//	glPopAttrib();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::bindTexture(sp<ImageI> spImageI)
{
	if(spImageI==m_currentTexture)
	{
		return;
	}

	m_currentTexture=spImageI;

	if(!spImageI.isValid())
	{
		useDefaultTexture();
		return;
	}

	GLenum format;
	GLenum type;
	GLint components;
	queryFormat(spImageI,format,type,components);

	const GLenum target=GL_TEXTURE_2D;
	const GLint levels=0;
	const GLint border=0;
	glTexImage2D(target,levels,components,
			(GLsizei)spImageI->width(),(GLsizei)spImageI->height(),border,
			format,type,(const GLvoid*)spImageI->raw());
}

} // namespace
