/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operate/operate.pmh"

using namespace fe;
void OperatorSurfaceCommon::bind(sp<SignalerI> a_spSignalerI,
									sp<Layout> a_spLayout)
{
	m_aStartFrame.initialize(a_spLayout->scope(),
			"StartFrame");
	m_aEndFrame.initialize(a_spLayout->scope(),
			"EndFrame");
	m_aFrame.initialize(a_spLayout->scope(),
			"Frame");
	m_aTime.initialize(a_spLayout->scope(),
			"Time");
	m_aSurfaceOutput.initialize(a_spLayout->scope(),
			"SurfaceOutput");
	m_aDrawOutput.initialize(a_spLayout->scope(),
			"DrawI");
	m_aDrawGuide.initialize(a_spLayout->scope(),
			"DrawGuide");
	m_aDrawBrush.initialize(a_spLayout->scope(),
			"DrawBrush");
	m_aDrawBrushOverlay.initialize(a_spLayout->scope(),
			"DrawBrushOverlay");
	m_aWindowEvent.initialize(a_spLayout->scope(),
			"WindowEvent");
	m_aCameraTransform.initialize(a_spLayout->scope(),
			"CameraTransform");
	m_aRayOrigin.initialize(a_spLayout->scope(),
			"RayOrigin");
	m_aRayDirection.initialize(a_spLayout->scope(),
			"RayDirection");
}

String OperatorSurfaceCommon::generateSignature(
	sp<SurfaceAccessibleI>& a_rspAccessibleI)
{
	sp<SurfaceAccessorI> spPoints;
	if(!access(spPoints,a_rspAccessibleI,e_point,e_generic))
	{
		return "";
	}

	sp<SurfaceAccessorI> spVertices;
	if(!access(spVertices,a_rspAccessibleI,e_primitive,e_vertices))
	{
		return "";
	}

	const I32 pointCount=spPoints->count();
	const I32 primitiveCount=spVertices->count();

	String signature;
	signature.sPrintf("%d:%d",pointCount,primitiveCount);

	return signature;
}

//* static
void OperatorSurfaceCommon::drawLabel(sp<DrawI> a_spDrawI,
	const SpatialVector& a_location,String a_label,
	U32 a_margin, const Color a_textColor,const Color* a_pBorderColor,
	const Color* a_pBackground)
{
	//* TODO query
	const U32 fontWidth=5;
	const U32 fontHeight=7;

	const SpatialVector behind=a_location-SpatialVector(0.0,0.0,0.5);

	SpatialVector rect[2];
	rect[0]=behind-SpatialVector(a_margin-1,a_margin);
	rect[1]=behind+SpatialVector(
			fontWidth*a_label.length()+a_margin,
			fontHeight+a_margin+1);

	Box2i backbox;
	set(backbox,rect[0][0]-2,rect[0][1]-2,
			rect[1][0]-rect[0][0]+4,rect[1][1]-rect[0][1]+4);

	if(a_pBackground)
	{
		a_spDrawI->drawRectangles(rect,2,false,a_pBackground);
	}

	if(a_pBorderColor)
	{
		a_spDrawI->drawBox(backbox,*a_pBorderColor);
	}

/*	text outline
	const Color black(0.0,0.0,0.0,1.0);
	const SpatialVector offsetX(1.0,0.0,0.0);
	const SpatialVector offsetY(0.0,1.0,0.0);
	a_spDrawI->drawAlignedText(a_location+offsetX,a_label,black);
	a_spDrawI->drawAlignedText(a_location-offsetX,a_label,black);
	a_spDrawI->drawAlignedText(a_location+offsetY,a_label,black);
	a_spDrawI->drawAlignedText(a_location-offsetY,a_label,black);
*/

	a_spDrawI->drawAlignedText(a_location,a_label,a_textColor);
}

//* static
void OperatorSurfaceCommon::drawPicker(sp<DrawI> a_spDrawI,
	const SpatialVector& a_point,const SpatialVector& a_up,
	const SpatialVector& a_normal,Real a_scale,I32 a_depth,
	const Color* a_pColor)
{
	SpatialTransform transform;
	makeFrameNormalY(transform,a_point,a_up,a_normal);
	rotate(transform,-90.0*degToRad,e_xAxis);

	SpatialVector scale(a_scale,a_scale,a_scale);
	a_spDrawI->drawCircle(transform,&scale,*a_pColor);

	for(I32 m=1;m<=a_depth;m++)
	{
		SpatialVector line[4];
		line[0]=SpatialVector(-1.0-0.5*m,-0.5);
		line[1]=SpatialVector(-1.0-0.5*m,0.5);
		line[2]=SpatialVector(1.0+0.5*m,-0.5);
		line[3]=SpatialVector(1.0+0.5*m,0.5);
		a_spDrawI->drawTransformedLines(transform,&scale,line,NULL,4,
				DrawI::e_discrete,false,a_pColor);
	}
}

//* static
void OperatorSurfaceCommon::drawImpact(sp<DrawI> a_spDrawI,
	sp<SurfaceTriangles::Impact> a_spTriImpact,
	Real a_inset,const Color* a_pColor)
{
	const Real scale=1.0-2.0*a_inset;

	SpatialVector line[4];
	line[0]=scale*a_spTriImpact->vertex0()+
			a_inset*a_spTriImpact->vertex1()+
			a_inset*a_spTriImpact->vertex2();
	line[1]=a_inset*a_spTriImpact->vertex0()+
			scale*a_spTriImpact->vertex1()+
			a_inset*a_spTriImpact->vertex2();
	line[2]=a_inset*a_spTriImpact->vertex0()+
			a_inset*a_spTriImpact->vertex1()+
			scale*a_spTriImpact->vertex2();
	line[3]=line[0];
	a_spDrawI->drawLines(line,NULL,4,DrawI::e_strip,false,a_pColor);
}

//* static
void OperatorSurfaceCommon::highlightCurve(sp<ViewI> a_spView,
	sp<DrawI> a_spDraw,sp<SurfaceAccessorI> a_spVertices,
	I32 a_primitiveIndex,I32 a_vertexIndex,
	const Color* a_pLineColor,const Color* a_pPointColor,
	const Color* a_pPointHighlight)
{
	const I32 subCount=a_spVertices->subCount(a_primitiveIndex);
	SpatialVector line[subCount];

	for(I32 subIndex=0;subIndex<subCount;subIndex++)
	{
		const SpatialVector point=
				a_spVertices->spatialVector(a_primitiveIndex,subIndex);

		const Vector2i projected=
				a_spView->project(point,ViewMode::e_perspective);

		line[subIndex]=projected;	//* NOTE z defaults to 0.0
	}

	if(a_pLineColor)
	{
		a_spDraw->drawLines(line,NULL,subCount,
				DrawI::e_strip,FALSE,a_pLineColor);
	}

	if(a_pPointColor)
	{
		for(I32 subIndex=0;subIndex<subCount;subIndex++)
		{
			line[subIndex][2]=0.5;
		}

		a_spDraw->drawPoints(line,NULL,subCount,FALSE,a_pPointColor);
	}

	if(a_vertexIndex>=0 && a_pPointHighlight)
	{
		const SpatialVector point=
				a_spVertices->spatialVector(a_primitiveIndex,a_vertexIndex);
		drawDot(a_spView,a_spDraw,point,5.0,*a_pPointHighlight);
	}
}

//* static
void OperatorSurfaceCommon::drawDot(sp<ViewI> a_spView,
	sp<DrawI> a_spDraw,SpatialVector a_point,Real a_radius,const Color a_color)
{
		const Vector2i projected=
				a_spView->project(a_point,ViewMode::e_perspective);
		SpatialVector center(projected[0],projected[1],0.6);

		SpatialTransform transform;
		setIdentity(transform);
		translate(transform,center);

		const SpatialVector scale(a_radius,a_radius,a_radius);
		a_spDraw->drawCircle(transform,&scale,a_color);
}

//* static
void OperatorSurfaceCommon::drawSquare(sp<DrawI>& a_rspDraw,
	const SpatialVector& a_rPoint,Real a_rimSize,
	const Color& a_rColorCenter,const Color& a_rColorRim,
	const Color* a_pColorShadow,const Color* a_pColorSparkle)
{
	sp<ViewI> spView=a_rspDraw->view();

	const Vector2i projected=spView->project(a_rPoint,ViewMode::e_perspective);
	const Vector2i rim(a_rimSize,a_rimSize);

	Box2i box=projected-rim;
	box.size()=2.0*rim+Vector2i(1,1);

	if(a_rColorCenter[3]>0.0)
	{
		const I32 lineWidth=a_rspDraw->drawMode()->lineWidth();

		SpatialVector corner[2];
		corner[0]=SpatialVector(box)+SpatialVector(lineWidth,lineWidth);
		corner[1]=SpatialVector(box)-SpatialVector(lineWidth,lineWidth)+
				SpatialVector(box.size());

		a_rspDraw->drawRectangles(corner,2,FALSE,&a_rColorCenter);
	}

	if(a_rColorRim[3]>0.0)
	{
		a_rspDraw->drawBox(box,a_rColorRim);
	}

	if(a_pColorShadow && (*a_pColorShadow)[3]>0.0)
	{
		SpatialVector glint=SpatialVector(box)+
				SpatialVector(box.size()[0],0.0,1.0);
		a_rspDraw->drawPoints(&glint,NULL,1,FALSE,a_pColorShadow);
	}

	if(a_pColorSparkle && (*a_pColorSparkle)[3]>0.0)
	{
		SpatialVector glint=SpatialVector(box)+
				SpatialVector(0.0,box.size()[1],1.0);
		a_rspDraw->drawPoints(&glint,NULL,1,FALSE,a_pColorSparkle);
	}
}


//* static
SpatialVector OperatorSurfaceCommon::closestVertex(
	sp<SurfaceTriangles::Impact>& a_rspTriImpact)
{
	const SpatialBary bary=a_rspTriImpact->barycenter();
	const SpatialVector bary3=bary;
	if(bary3[0]>bary3[1])
	{
		if(bary3[0]>bary3[2])
		{
			return a_rspTriImpact->vertex0();
		}
		else
		{
			return a_rspTriImpact->vertex2();
		}
	}
	else if(bary3[1]>bary3[2])
	{
		return a_rspTriImpact->vertex1();
	}
	return a_rspTriImpact->vertex2();
}

//* static
Real OperatorSurfaceCommon::nearestPointOnCurve(const SpatialVector a_point,
	const sp<SurfaceAccessorI> a_spCurves,const I32 a_index)
{
	const U32 subCount=a_spCurves->subCount(a_index);
	if(subCount<2)
	{
		return 0.0;
	}

	Real result=0.0;
	Real minDist=0.0;

	SpatialVector lastCV=a_spCurves->spatialVector(a_index,0);
	for(U32 subIndex=1;subIndex<subCount;subIndex++)
	{
		const SpatialVector cv=a_spCurves->spatialVector(a_index,subIndex);
		const SpatialVector segment=cv-lastCV;
		const Real segLength=magnitude(segment);
		const Real safeLength=segLength>0.0? segLength: 1.0;
		const SpatialVector unitSeg=segment/safeLength;
		const SpatialVector away=a_point-lastCV;
		const Real along=dot(unitSeg,away);
		const Real beyond=
				along>segLength? along-segLength: (along<0.0? -along: 0.0);
		const SpatialVector aside=away-along*unitSeg;
		const Real astray=magnitude(aside);
		const Real dist=sqrt(astray*astray+beyond*beyond);

		if(subIndex==1 || dist<minDist)
		{
			const Real unitAlong=along/safeLength;

			minDist=dist;
			result=subIndex-1+unitAlong;
		}

		lastCV=cv;
	}

	return result;
}
